#
# Copyright (C) 2020 Satoru SATOH <satoru.satoh@gmail.com>.
# SPDX-License-Identifier: MIT
#
r"""Parse fortios' `show *configuration' outputs and generates various outputs.

.. versionadded:: 0.1.0

   - initial checkin
"""
from __future__ import absolute_import

import collections.abc
import logging
import os.path
import re

import anyconfig

from . import utils


CNF_NAMES = ("system.*",
             "firewall service category",
             "firewall service group",
             "firewall service custom",
             "firewall addrgrp",
             "firewall address",
             "firewall policy")

METADATA_FILENAME = "metadata.json"
ALL_FILENAME = "all.json"

LOG = logging.getLogger(__name__)


def has_vdom(cnf):
    """
    .. note:: I guess that it's effective and faster than the following.

       ::

         return bool(jmespath.search("configs[?config=='vdom']", cnf))

    :param cnf: A mapping objects should have key: configs
    :return: True if vdoms are found in given configurations
    """
    return any(c for c in cnf.get("configs", []) if c.get("config") == "vdom")


def jmespath_search_1(path_exp, data, normalize_fn=None):
    """
    Return the result of JMESPath query to
    given data `data`.

    :param path_exp: JMESPath expression for search results
    :param data: A list of mapping objects

    :return: JMESPath search result :: [<mapping object>]
    """
    LOG.debug("%s: JMESPath exp: %s", __name__, path_exp)
    res = utils.search(path_exp, data)
    if callable(normalize_fn):
        if not utils.is_str(res) and isinstance(res, collections.Sequence):
            res = [normalize_fn(r) for r in res]
        else:
            res = normalize_fn(res)

    return res


def _global_path_exp(path_exp):
    """
    :param path_exp: JMESPath expression for search results
    :return: A JMESPath expression to search results

    >>> _global_path_exp("configs[?config=='system global']")
    "configs[?config=='global'] | [0].configs[?config=='system global']"
    """
    return "configs[?config=='global'] | [0]." + path_exp


def _vdoms_path_exp(path_exp, vdom=None):
    """
    :param path_exp: JMESPath expression for search results
    :param vdom: Specify vdom to search for

    :return: A JMESPath expression to search results

    >>> _vdoms_path_exp("configs[?config=='system global']")
    "configs[?config=='vdom'].edits[].configs[?config=='system global'][]"
    >>> _vdoms_path_exp("configs[]", vdom="root")
    "configs[?config=='vdom' && edits[0].edit=='root'].edits[].configs[]"
    """
    if vdom:
        return ("configs[?config=='vdom' && edits[0].edit=='{}']"
                ".edits[].".format(vdom) + path_exp)

    return "configs[?config=='vdom'].edits[]." + path_exp + "[]"


def jmespath_search(path_exp, data, normalize_fn=None, has_vdoms_=False,
                    vdom=None):
    """
    Similar to the above :func:`jmespath_search_1` but takes care the cases of
    configurations have multi vdoms.

    :param path_exp: JMESPath expression for search results
    :param data: A list of mapping objects
    :param has_vdoms_: True if givne `cnf` contains vdoms
    :param vdom: Specify vdom to make table

    :return: JMESPath search result :: [<mapping object>]
    """
    if not has_vdoms_:
        return jmespath_search_1(path_exp, data, normalize_fn=normalize_fn)

    # Both of the following results are type of [dict].
    res_global = jmespath_search_1(_global_path_exp(path_exp), data,
                                   normalize_fn=normalize_fn)
    res_vdoms = jmespath_search_1(_vdoms_path_exp(path_exp, vdom=vdom),
                                  data, normalize_fn=normalize_fn)
    if not res_global:
        return res_vdoms

    if not res_vdoms:
        return res_global

    return res_global + res_vdoms


def _config_edits_path_exp(cname):
    """
    :param cname: name of the configuration to search

    >>> _config_edits_path_exp("firewall address")
    "configs[?config=='firewall address'].edits[]"
    """
    return "configs[?config=='{}'].edits[]".format(cname)


def config_edits_search(cname, data, normalize_fn=None, has_vdoms_=False,
                        vdom=None):
    """
    Search configruations (edits).

    :param cname: Name of the configuration to search
    :param data: A list of mapping objects
    :param has_vdoms_: True if givne `cnf` contains vdoms
    :param vdom: Specify vdom to make table

    :return: JMESPath search result :: [<mapping object>]
    """
    pexp = _config_edits_path_exp(cname)
    return jmespath_search(pexp, data, normalize_fn=normalize_fn,
                           has_vdoms_=has_vdoms_, vdom=vdom)


def validate(cnf, filepath=None):
    """
    Validate `cnf` to fail first.

    :param cnf: Config data loaded or parsed log.
    :param filepath: File path gives config data `cnf`

    :raises: ValueError, TypeError
    """
    if not cnf:
        raise ValueError("No expected data was found in {}".format(filepath))

    if not isinstance(cnf, collections.abc.Mapping):
        raise TypeError("Invalid typed data was found in {}".format(filepath))

    if "configs" not in cnf:
        raise ValueError("Configs were not found in {}".format(filepath))

    if not isinstance(cnf["configs"], collections.abc.Iterable):
        raise ValueError("Configs :: [] was not found in {}".format(filepath))


def parse_show_config(filepath):
    """
    Parse 'show full-configuration output and returns a list of parsed configs.

    :param filepath:
        a str or :class:`pathlib.Path` object represents file path contains
        'show full-configuration` or 'show' output

    :return: A mapping object holding configurations
    :raises: IOError, OSError and so on
    """
    cnf = utils.try_ac_load(filepath, type_="fortios")
    validate(cnf, filepath)

    return cnf


def load(filepath):
    """
    Load JSON file contains parsed results

    :param filepath:
        a str or :class:`pathlib.Path` object represents file path contains
        parsed results of 'show full-configuration` or 'show' output

    :return: A mapping object holding configurations
    :raises: IOError, OSError and so on
    """
    cnf = utils.try_ac_load(filepath)
    validate(cnf, filepath)

    return cnf


def hostname_from_configs(cnf, has_vdoms_=False):
    """
    Detect hostname of the fortigate node from its 'system global'
    configuration.

    :param cnf: Config data loaded or parsed log.
    :param has_vdoms_: True if givne `cnf` contains vdoms

    :return: hostname str or None (if hostname was not found)
    """
    sgcnf = jmespath_search("configs[?config=='system global']", cnf,
                            has_vdoms_=has_vdoms_)

    if not sgcnf:  # I believe that it never happen.
        raise ValueError("No system global configs found. Is it correct data?")

    return sgcnf[0].get("hostname", '').lower() or None


def list_vdom_names(cnf):
    """
    Pattern:
        {"config": "vdom",
         "edits": [{"edit": "root"}, ...]}

    :param cnf: Config data loaded or parsed log.
    :param list_names: List vdoms with names

    :return: A list of the name of VDoms
    """
    if not has_vdom(cnf):
        return ["root"]

    vdoms = set(jmespath_search("configs[?config=='vdom'].edits[0].edit", cnf))
    if not vdoms:
        raise ValueError("VDoms were not found. Is it correct data?")

    return sorted(vdoms)


def unknown_name():
    """
    Compute the hostname using checksum of `inpath`
    """
    return "unknown-{}".format(utils.timestamp())


def cname_to_filename(cname, ext=".json"):
    """
    :return: A filename generated from `cname`
    """
    return re.sub(r"[\s\"']", '_', cname) + ext


def list_cnames_for_regexp(cnf, regexp=None, has_vdoms_=False):
    """List config names.

    :param cnf: Config data loaded or parsed log.
    :param has_vdoms_: True if givne `cnf` contains vdoms
    """
    return sorted(set(c for c in jmespath_search("configs[].config[]", cnf,
                                                 has_vdoms_=has_vdoms_)
                      if regexp.match(c)))


def parse_show_config_and_dump(inpath, outdir=None, cnames=CNF_NAMES):
    """
    :param inpath:
        a str or :class:`pathlib.Path` object represents file path contains
        'show full-configuration` or any other 'show ...' outputs
    :param outdir: Dir to save parsed results as JSON files

    :return: A mapping object contains parsed results
    :raises: IOError, OSError
    """
    cnf = parse_show_config(inpath)  # {"configs": [...]}

    vdoms = list_vdom_names(cnf)
    _has_vdoms = vdoms and len(vdoms) > 1

    try:
        # It should have this in most cases.
        hostname = hostname_from_configs(cnf, has_vdoms_=_has_vdoms)
    except ValueError as exc:
        LOG.warning("%r: %s\nCould not resovle hostname", exc, inpath)
        hostname = unknown_name()

    if not outdir:
        outdir = "out"

    houtdir = os.path.join(outdir, hostname)
    utils.ensure_dir_exists(houtdir)

    anyconfig.dump(cnf, os.path.join(houtdir, ALL_FILENAME))

    gmark = '*'
    opts = dict(has_vdoms_=_has_vdoms)
    for cname in cnames:
        if gmark in cname:
            cregexp = re.compile(cname)
            if cregexp:
                ccnames = list_cnames_for_regexp(cnf, regexp=cregexp,
                                                 has_vdoms_=_has_vdoms)

                for ccn in ccnames:
                    pexp = "configs[?config=='{}'].edits[]".format(ccn)
                    ccnf = jmespath_search(pexp, cnf, **opts)
                    ccname = cname_to_filename(ccn)
                    anyconfig.dump(ccnf, os.path.join(houtdir, ccname))
        else:
            # TODO: Save configs per global and VDoms?
            pexp = "configs[?config=='{}'].edits[]".format(cname)
            ccnf = jmespath_search(pexp, cnf, **opts)
            ccname = cname_to_filename(cname)
            anyconfig.dump(ccnf, os.path.join(houtdir, ccname))

    anyconfig.dump(dict(timestamp=utils.timestamp(), hostname=hostname,
                        vdoms=vdoms, origina_data=inpath),
                   os.path.join(houtdir, METADATA_FILENAME))

    return cnf


def parse_show_configs_and_dump_itr(inpaths, outdir=None, cnames=CNF_NAMES):
    """
    :param inpaths:
        Similar to `inpath` in :func:`parse_show_config_and_dump` but consists
        of mulitple paths
    :param outdir: Dir to save parsed results as JSON files

    :yield:
        A list of a tuple of (input file path, mapping object contains parsed
        results)
    :raises: IOError, OSError
    """
    for inpath in inpaths:
        yield (inpath,
               parse_show_config_and_dump(inpath, outdir, cnames=cnames))

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2020 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
# pylint: disable=missing-docstring,invalid-name
from __future__ import absolute_import

import fortios_xutils.firewall as TT
import fortios_xutils.parser as P
import tests.common as C


class TestCases_10(C.TestCase):

    def test_10_df_by_query__mzero(self):
        rdf = TT.df_by_query("configs[]", {})
        self.assertTrue(rdf.empty)

    def test_12_df_by_query__found(self):
        rdf = TT.df_by_query("configs[]", dict(configs=[dict(a=1)]))
        self.assertFalse(rdf.empty)
        self.assertEqual(rdf.to_dict(orient="record"), [dict(a=1)])

    def test_30_guess_filetype(self):
        self.assertEqual(TT.guess_filetype("foo.pickle"), "pickle")
        self.assertEqual(TT.guess_filetype("foo.json"), "json")
        self.assertEqual(TT.guess_filetype("foo.json.gz", compression=True),
                         "json")


class TestCases_20(C.TestCaseWithWorkdir):

    def test_10_pandas_save_load(self):
        for ext in "pickle json".split():
            filepath = TT.os.path.join(self.workdir, "1", "test." + ext)
            TT.pandas_save(TT.DF_ZERO, filepath)
            rdf = TT.pandas_load(filepath)
            self.assertTrue(rdf.empty)

            TT.pandas_save(TT.DF_ZERO, filepath + ".gz", compression="gzip")
            self.assertTrue(TT.os.path.exists(filepath + ".gz"))

    def test_12_pandas_save_load__excs(self):
        self.assertRaises(ValueError, TT.pandas_save, TT.DF_ZERO,
                          "/a/b/c.ext_not_exist")
        self.assertRaises(ValueError, TT.pandas_load, "/a/b/c.ext_not_exist")


class TestCasesWithConfigs(C.TestCase):

    cpaths = C.list_res_files("show_configs", "*.txt")


class TestCases_30(C.TestCaseWithWorkdir, TestCasesWithConfigs):

    def test_20_make_and_save_firewall_policy_table(self):
        for idx, cpath in enumerate(self.cpaths):
            opath = TT.os.path.join(self.workdir, str(idx), "fw.pickle")
            cnf = P.parse_show_config(cpath)
            rdf = TT.make_and_save_firewall_policy_table(cnf, opath)
            self.assertFalse(rdf.empty)
            self.assertTrue(TT.os.path.exists(opath))


class TestCases_40(TestCasesWithConfigs):

    def setUp(self):
        self.cnfs = [P.parse_show_config(p) for p in self.cpaths]

    def test_10_make_firewall_address_table_1(self):
        for cnf in self.cnfs:
            rdf = TT.make_firewall_address_table_1(
                cnf,
                has_vdoms_=P.has_vdom(cnf)
            )
            self.assertFalse(rdf.empty)
            self.assertTrue("SSLVPN_TUNNEL_ADDR1" in rdf["edit"].values)

    def test_20_make_firewall_addrgrp_table(self):
        for cnf in self.cnfs:
            rdf = TT.make_firewall_addrgrp_table(
                cnf,
                has_vdoms_=P.has_vdom(cnf)
            )
            self.assertFalse(rdf.empty)
            self.assertTrue("G Suite" in rdf["edit"].values)
            self.assertTrue("host_192.168.3.1" in rdf["member"].values)

    def test_30_make_firewall_address_table(self):
        for cnf in self.cnfs:
            rdf = TT.make_firewall_address_table(
                cnf,
                has_vdoms_=P.has_vdom(cnf)
            )
            self.assertFalse(rdf.empty)
            self.assertTrue("SSLVPN_TUNNEL_ADDR1" in rdf["edit"].values)
            self.assertTrue("G Suite" in rdf["edit"].values)
            self.assertTrue(any("192.168.1.0/24" in addrs
                                for addrs in rdf["addrs"].values))
            self.assertTrue(any("192.168.3.3/32" in addrs
                                for addrs in rdf["addrs"].values))


class TestCases_50(C.unittest.TestCase):

    def setUp(self):
        self.cpaths = C.list_res_files("show_configs", "*.txt")
        self.cnfs = [P.parse_show_config(p) for p in self.cpaths]
        self.fdfs = [
            TT.make_firewall_address_table(c, has_vdoms_=P.has_vdom(c))
            for c in self.cnfs
        ]
        self.pdfs = [TT.make_firewall_policy_table(c) for c in self.cnfs]

    def test_10_search_by_addr_1__fa_not_found(self):
        for fdf in self.fdfs:
            res = TT.search_by_addr_1("127.0.0.1", fdf)
            self.assertFalse(res)

    def test_12_search_by_addr_1__fa_found_1(self):
        for fdf in self.fdfs:
            res = TT.search_by_addr_1("192.168.5.201", fdf)
            self.assertTrue(res)
            self.assertEqual(len(res), 1)
            self.assertEqual(res[0]["edit"], "SSLVPN_TUNNEL_ADDR1")

    def test_20_search_by_addr_1__fp_not_found(self):
        for rdf in self.pdfs:
            res = TT.search_by_addr_1("127.0.0.1", rdf)
            self.assertFalse(res)

    def test_22_search_by_addr_1__fp_found_1(self):
        for rdf in self.pdfs:
            res = TT.search_by_addr_1("192.168.2.1", rdf)
            self.assertTrue(res)
            self.assertEqual(len(res), 1)
            self.assertEqual(res[0]["name"], "Monitor_Servers_01")

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2020 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
# pylint: disable=missing-docstring,invalid-name
from __future__ import absolute_import

import subprocess

import fortios_xutils.utils as TT
import tests.common as C


class SimpleFunctionTestCases(C.unittest.TestCase):

    maxDiff = None

    def test_20_timestamp(self):
        dt = TT.datetime.datetime(2020, 1, 7, 1, 23, 45)
        ref = "2020-01-07_01_23_45"
        self.assertEqual(TT.timestamp(dt), ref)

    def test_30_checksum(self):
        cs = "md5sum {}".format(__file__).split()
        ref = subprocess.check_output(cs).decode("utf-8").split()[0]

        self.assertEqual(TT.checksum(__file__), ref)

    # TODO:
    def test_50_try_ac_load__ng(self):
        pass

    def test_52_try_ac_load__ok(self):
        ok_files = C.list_res_files("*_ok.json")
        for f in ok_files:
            self.assertTrue(TT.try_ac_load(f) is not None)

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2020 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
# pylint: disable=missing-docstring,invalid-name
from __future__ import absolute_import

import functools

import fortios_xutils.finder as TT
import tests.common as C


# tests/res/networks/graph.yml
NET_CONF_PATH = C.list_res_files("networks", pattern="*.yml")[0]


class TestCases_10(C.unittest.TestCase):

    def test_10_validate(self):
        self.assertRaises(ValueError, TT.validate, {})
        self.assertRaises(TypeError, TT.validate, [1])
        self.assertRaises(ValueError, TT.validate, dict(a=1))
        self.assertRaises(ValueError, TT.validate, dict(nodes=1, links=[]))
        self.assertRaises(ValueError, TT.validate, dict(nodes=[], links=1))

    def test_20_load(self):
        self.assertTrue(TT.load(NET_CONF_PATH))

    def test_30_find_paths(self):
        graph = TT.load(NET_CONF_PATH)
        fnc = functools.partial(TT.find_paths, graph)

        pss = fnc("192.168.122.2", "192.168.5.10")
        self.assertTrue(pss)
        self.assertTrue(pss[0])
        self.assertTrue(n.get("id", '') == "192.168.122.0/24" for n in pss)
        self.assertTrue(n.get("id", '') == "192.168.5.0/24" for n in pss)

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2020 Satoru SATOH <satoru.satoh@gmail.com>.
# SPDX-License-Identifier: MIT
#
r"""Network relaated functions.
"""
from __future__ import absolute_import

import functools
import ipaddress
import math
import operator
import re

import netaddr

from . import utils


IPV4_IP_RE_S = r"^(\d{1,3}\.){3}\d{1,3}"
IPV4_IP_RE = re.compile(IPV4_IP_RE_S + r'$')
UNI_NETMASK_RE = re.compile(r"^255.255.255.255$")

NET_MAX_PREFIX = 24


@functools.lru_cache(maxsize=32)
def normalize_ip(ip_s, prefix="/32"):
    """
    >>> normalize_ip("192.168.122.1")
    '192.168.122.1/32'
    >>> normalize_ip("192.168.122.1/32")
    '192.168.122.1/32'
    >>> normalize_ip("192.168.122.0/24")
    '192.168.122.0/24'
    """
    if '/' not in ip_s:  # e.g. 192.168.122.1
        return ip_s + prefix  # Normalize it.

    return ip_s


@functools.lru_cache(maxsize=32)
def is_network_address(addr, sep='/'):
    """
    :param addr: IP address string with prefix, e.g. 192.168.1.0/24

    >>> is_network_address("10.0.0.1/32")
    False
    >>> is_network_address("10.0.0.0/8")
    True
    """
    if sep in addr:
        return addr.split(sep)[-1] != '32'

    return ipaddress.ip_network(addr).num_addresses > 1


@functools.lru_cache(maxsize=32)
def subnet_to_ip(addr, netmask):
    """
    Convert fortios 'subnet' (addr, netmask) to ipaddress object.

    :return: <ip_address_with_prefix :: str>, e.g. 192.168.1.0/24, 10.0.0.1/32
    """
    if not utils.is_str(addr) or not utils.is_str(netmask):
        raise ValueError("Should be str but: {!r}/{!r}".format(addr, netmask))

    if not IPV4_IP_RE.match(addr) or not IPV4_IP_RE.match(netmask):
        raise ValueError("Should be address but: "
                         "{!s}/{!s}".format(addr, netmask))

    if UNI_NETMASK_RE.match(netmask):  # Unicast (host) address
        return str(ipaddress.ip_interface(addr))

    try:
        return str(ipaddress.ip_network('/'.join((addr, netmask))))
    except ValueError:  # It should be a host address with mask other than /32.
        return str(ipaddress.ip_interface(addr))  # Ignore the original netmask


@functools.lru_cache(maxsize=32)
def iprange_to_ipsets(start_ip, end_ip, prefix=32):
    """
    Convert IP range {start_ip, end_ip} to IP sets [<ip_0>, ...]

    >>> iprange_to_ipsets("192.168.122.1", "192.168.122.3")
    ['192.168.122.1/32', '192.168.122.2/32', '192.168.122.3/32']
    >>> iprange_to_ipsets("192.168.122.1", "192.168.122.3", 24)
    ['192.168.122.1/24', '192.168.122.2/24', '192.168.122.3/24']
    """
    if not utils.is_str(start_ip) or not utils.is_str(end_ip):
        raise ValueError("Should be str but: {!r}/{!r}".format(start_ip,
                                                               end_ip))

    if not IPV4_IP_RE.match(start_ip) or not IPV4_IP_RE.match(end_ip):
        raise ValueError("Should be address but: "
                         "{}/{}".format(start_ip, end_ip))

    # start_ip and end_ip should be in different networks.
    if start_ip.split('.')[0] != end_ip.split('.')[0]:
        raise ValueError("Looks in different networks: "
                         "{}/{}".format(start_ip, end_ip))

    return ["{!s}/{!s}".format(ip, prefix)
            for ip in netaddr.iter_iprange(start_ip, end_ip)]


@functools.lru_cache(maxsize=32)
def to_network(addr_s):
    """
    :param addr_s:
        A str represents an any IP address, maybe a host (unicast) or network
        address
    :return: An IPv*Network object or None

    >>> to_network("10.0.1.0/24")
    IPv4Network('10.0.1.0/24')
    >>> to_network("10.0.1.2/32")
    IPv4Network('10.0.1.2/32')
    >>> to_network("10.0.1.2")
    IPv4Network('10.0.1.2/32')
    >>> to_network("10.0.1.2/24") is None
    True
    >>> to_network("aaa") is None
    True
    """
    if isinstance(addr_s, (ipaddress.IPv4Network, ipaddress.IPv6Network)):
        return addr_s  # Nothing to do.

    if not utils.is_str(addr_s):
        raise ValueError("A str is expected but not: {!r}".format(addr_s))

    try:
        return ipaddress.ip_network(addr_s)

    # addr_s is not an ip address or an address with host bit set, etc.
    except ValueError:
        return None


def to_networks(*addrs):
    """
    :param addrs:
        A list of a str represents an any IP address, maybe a host (unicast) or
        network address
    :return: A list of IPv*Network objects
    """
    return [n for n in (to_network(a) for a in addrs) if n]


def list_addrs_contain_the_ip_itr(ip_s, addrs):
    """
    :param ip_s: A str represents an (unicast, host) IP address, e.g. 10.1.1.1
    :param addrs:
        An iterable of each item represents a host or a network address, e.g.
        10.0.0.0/8, 192.168.122.1/32

    :return: A list of ip addresses of networks contain the ip `ip_s`

    >>> addrs = ("192.168.122.0/24", "192.168.1.0/24", "10.0.1.0/24",
    ...          "192.168.122.1/32", "10.0.1.254/32")
    >>> list(list_addrs_contain_the_ip_itr("192.168.10.254", addrs))
    []
    >>> list(list_addrs_contain_the_ip_itr("10.1.0.254", addrs))
    []
    >>> list(list_addrs_contain_the_ip_itr("192.168.122.1", addrs))
    ['192.168.122.0/24', '192.168.122.1/32']
    >>> list(list_addrs_contain_the_ip_itr("192.168.122.254", addrs))
    ['192.168.122.0/24']
    >>> list(list_addrs_contain_the_ip_itr("10.0.1.1", addrs))
    ['10.0.1.0/24']
    """
    if not addrs:
        return

    ipa = normalize_ip(ip_s)  # Add prefix if it's not set.

    for addr in addrs:
        if ipa == addr:  # Try exact match case first.
            yield addr
            continue

        # Test if the network `addr` contains the ip `ipa`.
        net = to_network(addr)
        if net is None:
            net = to_network(addr.split('/')[0])  # Strip prefix and ...

        if ipaddress.ip_interface(ipa) in net:
            yield addr


def is_ip_in_addrs(ip_s, addrs):
    """
    :param ip_s: A str represents an (unicast, host) IP address, e.g. 10.1.1.1
    :param addrs:
        An iterable of each item represents a host or a network address, e.g.
        10.0.0.0/8, 192.168.122.1/32

    :return: True if the network `net_s` contains the ip `ip_s`

    >>> addrs = ("192.168.122.0/24", "192.168.1.0/24", "10.0.1.0/24",
    ...          "192.168.122.1/32", "10.0.1.254/24")
    >>> is_ip_in_addrs("192.168.10.254", addrs)
    False
    >>> is_ip_in_addrs("10.1.0.254", addrs)
    False
    >>> is_ip_in_addrs("192.168.122.1", addrs)
    True
    >>> is_ip_in_addrs("192.168.122.254", addrs)
    True
    >>> is_ip_in_addrs("10.0.1.1", addrs)
    True
    """
    return any(list_addrs_contain_the_ip_itr(ip_s, addrs))


def _is_subnet_of(net1, net2):
    """An wrapper around ipaddress.IP*Network.subnet_of.

    >>> net1 = ipaddress.ip_network("192.168.122.0/24")
    >>> net2 = ipaddress.ip_network("192.168.0.0/16")
    >>> net3 = ipaddress.ip_network("192.168.1.0/24")
    >>> _is_subnet_of(net1, net1)
    True
    >>> _is_subnet_of(net1, net2)
    True
    >>> _is_subnet_of(net1, net3)
    False
    """
    try:
        return net1.subnet_of(net2)
    except AttributeError:  # ipaddress in py36 does not have the above.
        return (net1 == net2 or
                (net2.network_address <= net1.network_address and
                 net2.broadcast_address >= net1.broadcast_address))


def supernet_of_networks(*net_addrs, max_prefix=32):
    """
    Degenerate and summarize given network addresses. For example,

    :param net_addrs:
        A list of strings gives and ip address with prefix or
        ipaddress.IPv*Network objects
    :param max_prefix: Max prefix of candidate networks

    >>> net1 = "192.168.122.0/24"
    >>> net2 = "192.168.122.2/24"  # !net
    >>> net3 = ipaddress.ip_network("192.168.122.0/28")
    >>> net4 = ipaddress.ip_network("192.168.1.0/24")
    >>> net5 = ipaddress.ip_network("192.168.1.10/32")  # host
    >>> supernet_of_networks(net1, net2, net3)
    IPv4Network('192.168.122.0/24')
    >>> supernet_of_networks(net1, net2, net3, net4)
    IPv4Network('192.168.0.0/17')
    >>> supernet_of_networks(net1, net4, max_prefix=16)
    IPv4Network('192.168.0.0/16')
    >>> net5 = ipaddress.ip_network("10.1.0.0/16")
    >>> supernet_of_networks(net1, net5)
    >>> supernet_of_networks(net1, net5, max_prefix=1)
    """
    nets = sorted(to_networks(*net_addrs),
                  key=operator.attrgetter("prefixlen"))
    if not nets:
        return None

    max_prefix = min(nets[0].prefixlen, max_prefix)

    # try to find the "smallest" (broadest) network.
    for prefix in reversed(range(1, max_prefix + 1)):
        cnet = nets[0].supernet(new_prefix=prefix)
        if all(_is_subnet_of(n, cnet) for n in nets[1:]):
            return cnet

    return None


def distance(net1, net2, base=1):
    """
    Compute 'distance' between a pair of networks by these addresses only.

    :param net1: A str represents a network with prefix, e.g. 10.0.1.0/24
    :param net2: Likewise
    :return: An int gives 'distance' indicator between a pair of networks

    >>> net0 = "192.168.122.1"
    >>> net1 = "192.168.122.0/24"
    >>> net2 = "192.168.0.0/16"
    >>> net3 = "192.168.1.0/24"
    >>> net4 = "192.168.254.0/24"
    >>> net5 = "0.0.0.0/32"
    >>> distance(net1, net1)
    0
    >>> distance(net0, net1)
    1
    >>> distance(net1, net2)
    8
    >>> distance(net1, net3)
    14
    >>> distance(net3, net4)
    16
    >>> distance(net1, net5)
    inf
    """
    if net1 == net2:
        return 0

    no1 = ipaddress.ip_network(net1)
    no2 = ipaddress.ip_network(net2)

    # Case that either net1 or net2 is a host address in other network.
    if no1.num_addresses == 1 and ipaddress.ip_interface(net1) in no2:
        return base

    if no2.num_addresses == 1 and ipaddress.ip_interface(net2) in no1:
        return base

    # Case that either net1 or net2 contains other network.
    if _is_subnet_of(no1, no2) or _is_subnet_of(no2, no1):
        return base * abs(no1.prefixlen - no2.prefixlen)

    snet = supernet_of_networks(no1, no2)
    if snet is None:
        return math.inf

    return base * (no1.prefixlen + no2.prefixlen - 2 * snet.prefixlen)


def find_nearest_network(ipa, nets):
    """
    :param ipa: An ip address string
    :param nets:
        A of str gives and ip address with prefix, e.g. 10.0.1.0/24

    >>> net1 = "192.168.122.0/24"
    >>> net2 = "192.168.0.0/16"
    >>> net3 = "192.168.1.0/24"
    >>> net4 = "192.168.254.0/24"
    >>> net5 = "0.0.0.0/32"
    >>> find_nearest_network(net1, [net1, net5])
    '192.168.122.0/24'
    >>> find_nearest_network(net2, [net1, net5])
    '192.168.122.0/24'
    >>> find_nearest_network(net1, [net2, net3])
    '192.168.0.0/16'
    >>> find_nearest_network(net3, [net1, net4])
    '192.168.122.0/24'
    """
    return sorted(nets, key=functools.partial(distance, ipa))[0]

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2020 Satoru SATOH <satoru.satoh@gmail.com>.
# SPDX-License-Identifier: MIT
#
r"""Find network nodes.
"""
from __future__ import absolute_import

import collections.abc
import itertools

import networkx

from . import netutils, utils
from .network import NODE_ANY


def validate(cnf, filepath='N/A'):
    """
    Validate `cnf` to fail first.

    :param cnf: Config data loaded or parsed log.
    :param filepath: File path gives config data `cnf`

    :raises: ValueError, TypeError
    """
    if not cnf:
        raise ValueError("No expected data was found in {}".format(filepath))

    if not isinstance(cnf, collections.abc.Mapping):
        raise TypeError("Invalid typed data was found in {}".format(filepath))

    if any(k not in cnf for k in ("nodes", "links")):
        raise ValueError("Nodes and/or links were not found in "
                         "{}".format(filepath))

    if any(not isinstance(cnf[k], collections.abc.Iterable)
           for k in ("nodes", "links")):
        raise ValueError("Invalid types of nodes and links were "
                         "found in {}".format(filepath))


def load(filepath):
    """
    :param filepath:
        File path of network graph data ({'nodes': ..., 'links': ...})
    :param ac_args: keyword arguments given to anyconfig.load

    :return: An instance of networkx.Graph
    """
    cnf = utils.try_ac_load(filepath)
    validate(cnf, filepath)

    graph = networkx.Graph()
    graph.add_nodes_from((n["id"], n) for n in cnf["nodes"])
    for edge in cnf["links"]:
        graph.add_edge(edge["source"], edge["target"], **edge)

    return graph


def graph_nodes_itr(graph):
    """
    :param graph: A networkx.Graph object
    :yield: [{'addrs': [<ip_or_net_address_object>], ...}]
    """
    for idx in graph.nodes:
        yield graph.nodes[idx]


def find_nodes_by_ip_itr(nodes, ipa):
    """
    :param nodes: {'addrs': [<ip_network_address>], 'type': ..., ...}
    :param ipa: A str gives an ip address

    .. note:: 10.0.0.0/8 < 10.1.1.0/24 in ipaddress
    """
    for node in nodes:
        if netutils.is_ip_in_addrs(ipa, node.get("addrs", [])):
            yield node


def find_net_nodes_by_ip(nodes, ipa):
    """
    :param nodes: {'addrs': [<ip_network_address>], 'type': ..., ...}
    :param ipa: A str gives an ip address

    :return: [] or a list of network nodes sorted by its prefix

    .. note:: 10.0.0.0/8 < 10.1.1.0/24 in ipaddress
    """
    def _net_key_fun(node):
        """
        Compute the prefix length of the network `addr` to be used as key to
        sort network address list.
        """
        addr = node["addrs"][0]  # Net nodes should have an addr in addrs only.
        return netutils.to_network(addr).prefixlen

    return sorted(find_nodes_by_ip_itr(nodes, ipa), key=_net_key_fun,
                  reverse=True)


def find_a_net_node_by_ip(nodes, ipa):
    """
    :param nodes: {'addrs': [<ip_network_address>], 'type': ..., ...}
    :param ipa: A str gives an ip address

    :return: A network node or None

    .. seealso:: :func:`find_net_nodes_by_ip`
    """
    nets = find_net_nodes_by_ip(nodes, ipa)
    return nets[0] if nets else None


def select_unique_paths_itr(paths):
    """
    :param paths: A list of lists of nodes in the found paths
    :return: A generator yields a filtered a list of nodes in the paths
    """
    seen = set()
    for path in paths:  # path :: [{id: ..., } (node), ...]
        path_by_node_ids = tuple(n["id"] for n in path)
        if path_by_node_ids not in seen:
            seen.add(path_by_node_ids)
            yield path


def find_paths_itr(graph, src, dst, node_type=False, **nx_opts):
    """
    :param graph: networkx.Graph object to find paths from
    :param src: ipaddress.ip_address object or a string represents IP address
    :param dst: ipaddress.ip_address object or a string represents IP address
    :param node_type: Node type to filter results if given
    :param nx_opts:
        Keyword options given to networkx.all_shortest_paths() such as method,
        and weight

    :yield: A lists of nodes in the found paths
    :raises: ValueError if given src and/or dst is not an IP address string
    """
    (nodes_1, nodes_2) = itertools.tee(graph_nodes_itr(graph))

    src_net = find_a_net_node_by_ip(nodes_1, src)
    dst_net = find_a_net_node_by_ip(nodes_2, dst)

    if not src_net or not dst_net:
        return

    if src_net == dst_net:
        if not node_type or (node_type and node_type != NODE_ANY and
                             src_net["type"] == node_type):
            yield [src_net]

        return

    nss = networkx.all_shortest_paths(graph, src_net["id"], dst_net["id"],
                                      **nx_opts)
    res = [[n for n in graph_nodes_itr(graph) if n["id"] in ns] for ns in nss]

    if node_type and node_type != NODE_ANY:
        # Those paths might be degenerated and need to remove duplicates.
        pitr = ([n for n in ns if n["type"] == node_type] for ns in res)
        res = list(select_unique_paths_itr(pitr))

    for npath in res:
        # yield [src_net] + npath + [dst_net]
        yield npath


def find_paths(graph, src, dst, node_type=False, **nx_opts):
    """
    :param graph: networkx.Graph object to find paths from
    :param src: ipaddress.ip_address object or a string represents IP address
    :param dst: ipaddress.ip_address object or a string represents IP address
    :param node_type: Node type to filter results if given
    :param nx_opts:
        Keyword options given to networkx.all_shortest_paths() such as method,
        and weight

    :yield: A lists of nodes in the found paths
    :raises: ValueError if given src and/or dst is not an IP address string
    """
    return list(find_paths_itr(graph, src, dst, node_type=node_type,
                               **nx_opts))

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2020 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
# pylint: disable=missing-docstring,invalid-name
from __future__ import absolute_import

import fortios_xutils.netutils as TT
import tests.common as C


class TestCase_10_functions(C.unittest.TestCase):

    def test_20_subnet_to_ip__ng(self):
        self.assertRaises(ValueError, TT.subnet_to_ip, 1, 2)
        self.assertRaises(ValueError, TT.subnet_to_ip, "10.0.0.1", 8)
        self.assertRaises(ValueError, TT.subnet_to_ip, "10.0.0.1", "8")

    def test_22_subnet_to_ip__ok(self):
        self.assertEqual(TT.subnet_to_ip("10.0.0.0", "255.0.0.0"),
                         "10.0.0.0/8")
        self.assertEqual(TT.subnet_to_ip("10.0.0.1", "255.255.255.255"),
                         "10.0.0.1/32")
        self.assertEqual(TT.subnet_to_ip("10.0.0.1", "255.0.0.0"),
                         "10.0.0.1/32")

    def test_30_iprange_to_ipsets__ng(self):
        self.assertRaises(ValueError, TT.iprange_to_ipsets, 1, 2)
        self.assertRaises(ValueError, TT.iprange_to_ipsets, "10.0.0.1", 1)
        self.assertRaises(ValueError, TT.iprange_to_ipsets, 1, "10.0.0.1")
        self.assertRaises(ValueError,
                          TT.iprange_to_ipsets, "10.0.0.1", "192.168.1.1")

    def test_32_iprange_to_ipsets__ok(self):
        self.assertEqual(TT.iprange_to_ipsets("10.0.0.1", "10.0.0.3"),
                         ["10.0.0.1/32", "10.0.0.2/32", "10.0.0.3/32"])

    def test_50_is_ip_in_addrs__ng(self):
        ngs = [("10.1.1.1", []),
               ("10.1.1.1", ["192.168.1.0/24"]),
               ("10.0.0.1", ["10.1.0.0/16"]),
               ("10.0.0.0", ["10.1.0.0/16"])]

        for ip_s, nets in ngs:
            self.assertFalse(TT.is_ip_in_addrs(ip_s, nets))

    def test_52_is_ip_in_addrs__ok(self):
        ngs = [("10.1.1.1", ["10.1.1.1/32"]),
               ("10.1.1.1", ["192.168.1.0/24", "10.1.1.1/32"]),
               ("10.0.0.1", ["10.0.0.0/8"]),
               ("10.0.0.0", ["10.0.0.0/8"])]

        for ip_s, nets in ngs:
            self.assertTrue(TT.is_ip_in_addrs(ip_s, nets))

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2020 Satoru SATOH <satoru.satoh@gmail.com>.
# SPDX-License-Identifier: MIT
#
r"""Misc CLI commands.

.. versionadded:: 0.1.0

   - initial checkin
"""
from __future__ import absolute_import

import glob
import logging
import os.path

import anyconfig
import click

from fortios_xutils import finder, firewall, network, parser


LOG = logging.getLogger("fortios_xutils")


def expand_glob_paths_itr(filepaths):
    """
    :param filepaths: A list of file paths
    """
    for fpath in filepaths:
        if '*' in fpath:
            for path in sorted(glob.glob(fpath)):
                yield path
        else:
            yield fpath


@click.command()
@click.argument("filepaths", nargs=-1,
                type=click.Path(exists=True, readable=True))
@click.option("-O", "--outdir",
              help=("Output dir to save parsed results [out/ relative to "
                    "input filepath]"), default=None)
def parse(filepaths, outdir):
    """
    Parse fortigate CLI's "show *configuration* outputs and generate a
    structured JSON file. FILEPATHS is a list of file paths or a glob pattern
    gives that.

    Examples:

    \b
        $ fortios_xutils parse -O /tmp/0 \\
        > tests/res/show_configs/fortigate_cli_show_sample_*.txt
        $ ls /tmp/0
        fortigate-01  fortigate-02
        $ ls /tmp/0/fortigate-01:
        all.json                         firewall_address.json
        firewall_addrgrp.json            firewall_policy.json
        firewall_service_category.json   firewall_service_custom.json
        firewall_service_group.json      metadata.json
        system_global.json               system_interface.json
        system_object-tagging.json       system_replacemsg-group.json
        system_settings.json
        $ jq '.' /tmp/0/fortigate-01/system_interface.json
        [
          {
            "edit": "dmz",
            "vdom": "root",
            "status": "down",
            "type": "physical",
            "role": "dmz",
            "snmp-index": "1"
          },
          {
            "edit": "port1",
            "vdom": "root",
            "ip": [
              "192.168.122.10",
              "255.255.255.0"
            ],
                ... (snip) ...
    \f
    :param filepaths:
        A list of path of the input fortios' "show *configuration" output

    :param outdir: Dir to save parsed results as JSON files
    """
    fsit = expand_glob_paths_itr(filepaths)
    list(parser.parse_show_configs_and_dump_itr(fsit, outdir))


def parse_json_files_itr(filepaths, path_exp):
    """
    :param filepaths:
        A list of the JSON input file paths. Each JSON file contains parsed
        results
    :param path_exp: JMESPath expression to search for
    """
    for filepath in filepaths:
        cnf = parser.load(filepath)
        res = parser.jmespath_search(path_exp, cnf,
                                     has_vdoms_=parser.has_vdom(cnf))
        yield (filepath, res)


@click.command()
@click.argument("filepaths", nargs=-1,
                type=click.Path(exists=True, readable=True))
@click.option("-P", "--path", "path_exp", help="JMESPath expression to query")
def search(filepaths, path_exp):
    """
    Search an item or items from JSON file generated previously by 'parse' sub
    command. FILEPATHS is a list of file paths or a glob pattern gives that.

    Examples:

    \b
        $ # List ip addresses of system interfaces.
        $ fortios_xutils search \\
        > -P "configs[?config=='system interface'].edits[].ip" \\
        > tests/res/parsed/fortigate-01/all.json
        [
            [
                "192.168.122.10",
                "255.255.255.0"
            ],
            [
                "192.168.1.10",
                "255.255.255.0"
            ]
        ]
        $
    \f
    :param filepaths:
        A list of the JSON input file paths. Each JSON file contains parsed
        results
    :param path_exp: JMESPath expression to search for
    """
    fp_res_pairs = list(parse_json_files_itr(filepaths, path_exp))

    if len(filepaths) == 1:
        print(anyconfig.dumps(fp_res_pairs[0][1], ac_parser="json", indent=2))
    else:
        res = [dict(filepath=t[0], results=t[1]) for t in fp_res_pairs]
        print(anyconfig.dumps(res, ac_parser="json", indent=2))


@click.command()
@click.argument("filepaths", nargs=-1,
                type=click.Path(exists=True, readable=True))
@click.option("-P", "--prefix", help="Max network prefix [24]", default=24)
def network_collect(filepaths, prefix):
    """
    Make and save network data collected from the JSON structured fortigate's
    configuration files. FILEPATHS is a list of path of the JSON file, the
    parsed results of fortigate CLI's "show \\*configuration" outputs.

    Examples:

    \b
        $ fortios_xutils network-collect tests/res/parsed/*/all.json
        $ head -n 10 tests/res/parsed/fortigate-01/networks.yml
        metadata:
          type: metadata
          input: tests/res/parsed/fortigate-01/all.json
          prefix: 24
          timestamp: 2020-05-12_04_58_57
          version: '1.0'
        nodes:
        - id: fortigate-01
          name: fortigate-01
          type: firewall
        $

    \f
    :param filepaths:
        A list of path of the input JSON file which is the parsed results of
        fortios' "show *configuration" outputs
    :param prefix: Max network prefix to search networks for
    """
    fpaths = list(expand_glob_paths_itr(filepaths))
    list(network.make_and_save_networks_from_config_files_itr(fpaths,
                                                              prefix=prefix))


@click.command()
@click.argument("filepaths", nargs=-1,
                type=click.Path(exists=True, readable=True))
@click.option("-o", "--outpath",
              help="Path of the outpath file to save network JSON data",
              default=None)
def network_compose(filepaths, outpath):
    """
    Compose network files collected from the fortigate CLI's configurations
    from multiple fortigate hosts using the preivous network-collect command,
    into a network file.

    Examples:

    \b
        $ fortios_xutils network-compose \\
        > tests/res/parsed/fortigate-0*/networks.yml \\
        > -o tests/res/networks/all.yml
        $ head -n 10 tests/res/networks/all.yml
        metadata:
          inputs:
          - tests/res/parsed/fortigate-01/all.json
          - tests/res/parsed/fortigate-02/all.json
          timestamp: 2020-05-12_05_02_49
          version: '1.0'
        nodes:
        - id: fortigate-01
          name: fortigate-01
          type: firewall
        $

    \f
    :param filepaths:
        A list of network graph data files generated by `network_collect` sub
        command in advance.
    :param outpath: Path of the file to save data
    """
    fpaths = list(expand_glob_paths_itr(filepaths))
    network.compose_network_graph_files(fpaths, outpath=outpath)


@click.command()
@click.argument("filepath", type=click.Path(exists=True, readable=True))
@click.option("-o", "--outpath",
              help="Path of the outpath file to save pandas.DataFrame data",
              default=None)
def firewall_policy_save(filepath, outpath):
    """
    Make and save firewall policy table (:class:`pandas.DataFrame` object).

    Examples:

    \b
        $ fortios_xutils firewall-policy-save \\
        > -o /tmp/0/test.pickle.gz \\
        > tests/res/parsed/fortigate-01/all.json
        $ file /tmp/0/test.pickle.gz
        /tmp/0/test.pickle.gz: gzip compressed data, was "test.pickle"  ...
        $

    \f
    :param filepath:
        Path of the input JSON file which is the parsed results of fortios'
        "show *configuration" outpath
    :param outpath: Path of the file to save data
    """
    if not outpath:
        outpath = os.path.join(os.path.dirname(filepath), "out",
                               "result.pickle.gz")

    cnf = parser.load(filepath)
    firewall.make_and_save_firewall_policy_table(cnf, outpath,
                                                 compression="gzip")


@click.command()
@click.argument("filepath", type=click.Path(exists=True, readable=True))
@click.option("-i", "--ip", "ip_s",
              help="Specify an IP address to search")
def firewall_policy_search(filepath, ip_s):
    """
    Search firewall policy table generated by 'firewall-policy-save' command,
    by ip address. FILEPATH is a file path to the pandas dataframe file
    generated by 'firewall-policy-save' command.

    Examples:

    \b
        $ fortios_xutils firewall-policy-search \\
        > --ip 192.168.122.3 /tmp/0/test.pickle.gz
        [
          {
            "edit": "20",
            "name": "Monitor_Servers_02",
            "uuid": "3da73baa-dacb-48cb-852c-c4be245b4609",
            "srcintf": "port1",
            "dstintf": "",
            "srcaddr": "host_192.168.122.1",
            "dstaddr": "network_192.168.122.0/24",
            "action": "accept",
            "schedule": "always",
            "service": [
              "HTTPS",
              "HTTP"
            ],
            "inspection-mode": "",
            "nat": "",
            "srcaddrs": [
              "192.168.122.1/32"
            ],
            "dstaddrs": [
              "192.168.122.0/24"
            ],
            "comments": ""
          }
        ]

    \f
    :param filepath:
        Path of the json file contains parsed results of fortios' "show
        *configuration" outputs, or pandas.DataFrame data file, or the
        serialized pandas.DataFrame object contains firewall policy table
    :param ip_s: IP address string to search
    """
    if filepath.endswith("json"):
        cnf = parser.load(filepath)
        rdf = firewall.make_firewall_policy_table(cnf)
    else:
        rdf = firewall.pandas_load(filepath, compression="gzip")

    res = firewall.search_by_addr_1(ip_s, rdf)

    aopts = dict(ac_parser="json", indent=2)
    print(anyconfig.dumps(res, **aopts))


@click.command()
@click.argument("filepath", type=click.Path(exists=True, readable=True))
@click.argument("src_ip")
@click.argument("dst_ip")
@click.option("-N", "--ntype",
              help="Specify node type from the list: "
                   "{}".format(", ".join(network.NODE_TYPES)))
def network_find_paths(filepath, src_ip, dst_ip, ntype=None):
    """
    Search paths from the source `src_ip` to the destination `dst_ip`.

    Examples:

    \b
        $ fortios_xutils/cli.py network-find-paths \\
        > tests/res/networks/graph.yml 192.168.122.2 192.168.5.10
        [
          [
            {
              "id": "fortigate-01",
              "name": "fortigate-01",
              "type": "firewall",
              "addrs": [
                "192.168.122.10/24",
                "192.168.1.10/24"
              ],
              "type_id": 5,
              "class": "node firewall",
              "label": "fortigate-01 firewall"
            },
            {
              "id": "192.168.122.0/24",
              "name": "192.168.122.0/24",
              "type": "network",
              "addrs": [
                "192.168.122.0/24"
              ],
              "type_id": 1,
              "class": "node network",
              "label": "192.168.122.0/24 network"
            },
                ... (snip) ...
            {
              "id": "192.168.5.0/24",
              "name": "192.168.5.0/24",
              "type": "network",
              "addrs": [
                "192.168.5.0/24"
              ],
              "type_id": 1,
              "class": "node network",
              "label": "192.168.5.0/24 network"
            }
          ],
            ... (snip) ...
        $

    \f
    :param filepath:
        Path of the json or yaml file contains netowrk nodes and links
        information.
    :param src_ip: IP address of the source
    :param dst_ip: IP address of the destination
    """
    graph = finder.load(filepath)
    res = finder.find_paths(graph, src_ip, dst_ip)

    aopts = dict(ac_parser="json", indent=2)
    print(anyconfig.dumps(res, **aopts))


@click.group()
@click.option("-v", "--verbose", count=True, default=0)
def main(verbose=0):
    """CLI frontend entrypoint.
    """
    verbose = min(verbose, 2)
    LOG.setLevel([logging.WARNING, logging.INFO, logging.DEBUG][verbose])


for cmd in (parse, search, network_collect, network_compose,
            firewall_policy_save, firewall_policy_search,
            network_find_paths):
    main.add_command(cmd)

if __name__ == '__main__':
    main()

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2020 Satoru SATOH <satoru.satoh@gmail.com>.
# SPDX-License-Identifier: MIT
#
"""Common test module
"""
import glob
import os.path
import os
import shutil
import tempfile
import unittest

try:
    from unittest import SkipTest
except ImportError:
    from nose.plugins.skip import SkipTest


def skip_test():
    """Skip test cases
    """
    raise SkipTest


def selfdir(self=__file__):
    """
    >>> os.path.exists(selfdir())
    True
    """
    return os.path.dirname(self)


def resdir(self=__file__):
    """
    >>> assert os.path.exists(resdir())
    """
    return os.path.join(selfdir(self), "res")


def list_res_files(subdir, pattern="*"):
    """List resource data files
    """
    return sorted(glob.glob(os.path.join(resdir(), subdir, pattern)))


def setup_workdir():
    """Setup working dir
    """
    return tempfile.mkdtemp()


def prune_workdir(workdir):
    """Remove given `workdir` entirely.
    """
    shutil.rmtree(workdir)


class TestCase(unittest.TestCase):
    """Base class for test cases.
    """
    maxDiff = None


class TestCaseWithWorkdir(TestCase):
    """Base class for test cases need working dir.
    """
    cleanup = True
    workdir = None

    def setUp(self):
        """Prepare working dir
        """
        self.workdir = setup_workdir()

    def tearDown(self):
        """Cleanup working dir as needed
        """
        if self.cleanup:
            prune_workdir(self.workdir)

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2020 Satoru SATOH <satoru.satoh@gmail.com>.
# SPDX-License-Identifier: MIT
#
"""Utility functions
"""
from __future__ import absolute_import

import collections.abc
import datetime
import hashlib
import os.path
import os

import anyconfig
import jmespath


_ENCODINGS = ("utf-8", "shift-jis")


def ensure_dir_exists(filepath, subdir=None):
    """Ensure dir for given file path `filepath` exists

    :param filepath: File path might be created later
    :param subdir: Sub dir
    :return: A dir path
    """
    tdir = os.path.dirname(filepath)
    if subdir:
        tdir = os.path.join(tdir, subdir)

    if not os.path.exists(tdir):
        os.makedirs(tdir)

    return tdir


def is_str(obj):
    """
    >>> is_str(u"foo")
    True
    >>> is_str("192.168.122.1/24")
    True
    >>> import ipaddress
    >>> is_str(ipaddress.ip_interface("192.168.122.1/24"))
    False
    """
    return isinstance(obj, (str, collections.abc.ByteString))


def timestamp(date_=False):
    """Generate timestamp string.

    :param date_: datetime.datetime object or False (default)
    """
    return (date_ or datetime.datetime.now()).strftime("%F_%H_%M_%S")


def checksum(filepath, hcnstrctr=hashlib.md5, enc="utf-8"):
    """Compute the checksum of given file, `filepath`

    :raises: OSError, IOError
    """
    return hcnstrctr(open(filepath).read().encode(enc)).hexdigest()


def try_ac_load(filepath, type_=None, encodings=_ENCODINGS):
    """
    Try to open and load `filepath` using anyconfig.load.

    :param filepath: File path to load
    :param type_: File type of the file, `filepath`
    :param encodings: Character set encodings to try

    :return:
        A mapping object or sequence objects loaded from `filepath` or None
    """
    for enc in encodings:
        try:
            with open(filepath, encoding=enc) as inp:
                return anyconfig.load(inp, ac_parser=type_)
        except UnicodeDecodeError:
            pass

    return None


def search(jmespath_exp, data):
    """
    Just an wrapper for :func:`jmespath.search`

    >>> search("[].a[]",
    ...        [dict(a=[1, 2], b=2), dict(b=3, c=4), dict(a=[3,4])])
    [1, 2, 3, 4]
    >>> search("[].x", [dict(a=[1, 2], b=2)])
    []
    """
    return jmespath.search(jmespath_exp, data)

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2020 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
# pylint: disable=missing-docstring,invalid-name
from __future__ import absolute_import

import os.path

import fortios_xutils.network as TT
import fortios_xutils.parser as P
import tests.common as C


class Base(C.TestCase):

    cpaths = C.list_res_files("show_configs", "*.txt")
    cnfs = [P.parse_show_config(p) for p in cpaths]
    hnames = [P.hostname_from_configs(c, P.has_vdom(c)) for c in cnfs]
    sargss = [dict(has_vdoms_=P.has_vdom(c)) for c in cnfs]

    # .. seealso:: tests/res/show_configs/*.txt
    ref_ips = [TT.ipaddress.ip_interface("192.168.122.10/24"),
               TT.ipaddress.ip_interface("192.168.1.10/24")]
    ref_fas = set(("0.0.0.0/32",
                   "192.168.3.3/32",
                   "192.168.3.1/32",
                   "192.168.3.5/32",
                   "192.168.122.0/24",
                   "192.168.122.1/32",
                   "192.168.1.0/24",
                   "192.168.2.3/32",
                   "192.168.2.2/32",
                   "192.168.5.202/32",
                   "192.168.2.1/32",
                   "192.168.5.201/32",
                   "192.168.5.200/32"))


class TestCases_10(Base):

    def test_10_list_interface_addrs__no_data(self):
        for cnf, sargs in [({}, {})]:
            res = TT.list_interface_addrs(cnf, **sargs)
            self.assertFalse(res)

    def test_12_list_interface_addrs__found(self):
        for cnf, sargs in zip(self.cnfs, self.sargss):
            res = TT.list_interface_addrs(cnf, **sargs)
            self.assertTrue(res)
            self.assertEqual(res, self.ref_ips)

    def test_30_list_firewall_addrs__no_data(self):
        for cnf, sargs in [({}, {})]:
            res = TT.list_firewall_addrs(cnf, **sargs)
            self.assertFalse(res)

    def test_32_list_firewall_addrs(self):
        for cnf, sargs in zip(self.cnfs, self.sargss):
            res = TT.list_firewall_addrs(cnf, **sargs)
            self.assertTrue(res)
            self.assertEqual(set(res), self.ref_fas)


NG_CNF_S_1 = """
{"configs": [{"config": "system global", "hostname": "foo-1"}]}
"""

NG_CNF_S_2 = """
{"configs": [
    {"config": "global",
     "configs": [
        {"config": "system global",
         "hostname": "fortigate-02"}
     ]
    }
  ]
}
"""


def node_and_edges_from_config_file(fpath):
    return list(TT.node_and_edges_from_config_file_itr(fpath))


class TestCases_20(C.TestCaseWithWorkdir, Base):

    def test_10_node_and_edges_from_config_file_itr__wrong_data(self):
        cpath = os.path.join(self.workdir, "test.json")
        open(cpath, 'w').write("{}")
        self.assertRaises(ValueError, node_and_edges_from_config_file, cpath)

        open(cpath, 'w').write(NG_CNF_S_1)
        self.assertRaises(ValueError, node_and_edges_from_config_file, cpath)

        open(cpath, 'w').write(NG_CNF_S_2)
        self.assertRaises(ValueError, node_and_edges_from_config_file, cpath)

    def test_12_node_and_edges_from_config_file_itr(self):
        for (hname, cnf) in zip(self.hnames, self.cnfs):
            cpath = os.path.join(self.workdir, hname, "test.json")
            TT.utils.ensure_dir_exists(cpath)
            TT.anyconfig.dump(cnf, cpath)

            res = node_and_edges_from_config_file(cpath)
            self.assertTrue(res)
            self.assertTrue(any(x for x in res if x.get("type", "firewall")))
            self.assertTrue(any(x for x in res if x.get("type", "network")))
            self.assertTrue(any(x for x in res if x.get("type", "edge")))

    def test_20_make_and_save_networks_from_config_file__no_data(self):
        cpath = os.path.join(self.workdir, "test.json")
        fun = TT.make_and_save_networks_from_config_file

        open(cpath, 'w').write("{}")
        self.assertRaises(ValueError, fun, cpath)

        open(cpath, 'w').write(NG_CNF_S_1)
        self.assertRaises(ValueError, fun, cpath)

        open(cpath, 'w').write(NG_CNF_S_2)
        self.assertRaises(ValueError, fun, cpath)

    def test_22_make_and_save_networks_from_config_file(self):
        for (hname, cnf) in zip(self.hnames, self.cnfs):
            cpath = os.path.join(self.workdir, hname, "test.json")
            TT.utils.ensure_dir_exists(cpath)
            TT.anyconfig.dump(cnf, cpath)

            res = TT.make_and_save_networks_from_config_file(cpath)
            self.assertTrue(res)

    def test_30_make_and_save_networks_from_config_files(self):
        cpaths = [os.path.join(self.workdir, hname, "test.json")
                  for hname in self.hnames]

        for idx, cnf in enumerate(self.cnfs):
            TT.utils.ensure_dir_exists(cpaths[idx])
            TT.anyconfig.dump(cnf, cpaths[idx])

        fun = TT.make_and_save_networks_from_config_files_itr
        for fpath, res in fun(cpaths):
            self.assertTrue(res)

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2020 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
# pylint: disable=missing-docstring,invalid-name
from __future__ import absolute_import

import os.path

import fortios_xutils.parser as TT
import tests.common as C


class TestCases_10(C.unittest.TestCase):

    def test_10_has_vdom(self):
        cnf = dict(configs=[dict(config="vdom")])
        self.assertTrue(TT.has_vdom(cnf))
        self.assertFalse(TT.has_vdom({}))

    def test_20_validate(self):
        self.assertRaises(ValueError, TT.validate, {})
        self.assertRaises(TypeError, TT.validate, [1])
        self.assertRaises(ValueError, TT.validate, dict(a=1))
        self.assertRaises(ValueError, TT.validate, dict(configs=True))


class TestCases_20_jmespath_search(C.unittest.TestCase):

    maxDiff = None

    def test_20_jmespath_search_1(self):
        cnf = dict(a=dict(b=[dict(c="D")]))

        self.assertEqual(TT.jmespath_search_1("a.b[?c=='D'] | [0].c", cnf),
                         'D')

        self.assertEqual(TT.jmespath_search_1("a.b[?c=='D'] | [0].c", cnf,
                                              normalize_fn=str.lower),
                         'd')

    def test_30_jmespath_search__no_vdoms(self):
        cnf = dict(a=dict(b=[dict(c="D")]))
        pexp = "a.b[?c=='D'] | [0].c"

        ref = TT.jmespath_search_1(pexp, cnf)
        res = TT.jmespath_search(pexp, cnf)

        self.assertTrue(res)
        self.assertEqual(res, ref)

    def test_32_jmespath_search__w_vdoms_global_only(self):
        val = "foo-0"
        cnf = dict(configs=[dict(config="global",
                                 configs=[dict(config="system global",
                                               hostname=val)])])
        pexp = "configs[?config=='system global'] | [].hostname"

        res = TT.jmespath_search(pexp, cnf, has_vdoms_=True)

        self.assertTrue(res)
        self.assertEqual(res, [val])

    def test_34_jmespath_search__w_vdoms_vdoms_only(self):
        vals = ["foo-0", "bar-1"]
        scnf = dict(config="system global", hostname=vals[0])
        scnf_1 = dict(config="system global", hostname=vals[1])
        cnf = dict(configs=[dict(config="global", configs=[]),
                            dict(config="vdom",
                                 edits=[dict(edit="root",
                                             configs=[scnf]),
                                        dict(edit="mng",
                                             configs=[scnf_1])])])

        pexp = "configs[?config=='system global'] | [].hostname[]"

        res = TT.jmespath_search(pexp, cnf, has_vdoms_=True)

        self.assertTrue(res)
        self.assertEqual(res, vals)

    def test_36_jmespath_search__w_vdoms__both_global_and_vdoms(self):
        val = "foo-0"
        scnf = dict(config="system global", hostname=val)
        cnf = dict(configs=[dict(config="global",
                                 configs=[scnf]),
                            dict(config="vdom",
                                 edits=[dict(edit="root",
                                             configs=[scnf])])])

        pexp = "configs[?config=='system global'] | [].hostname[]"

        res = TT.jmespath_search(pexp, cnf, has_vdoms_=True)

        self.assertTrue(res)
        self.assertEqual(res, [val, val])

    def test_38_jmespath_search__w_vdoms_the_vdom_only(self):
        val = "foo-0"
        scnf = dict(config="system global", hostname=val)
        scnf_1 = dict(config="system global", hostname="bar-1")

        cnf = dict(configs=[dict(config="global", configs=[]),
                            dict(config="vdom",
                                 edits=[dict(edit="root",
                                             configs=[scnf])]),
                            dict(config="vdom",
                                 edits=[dict(edit="mng",
                                             configs=[scnf_1])])])

        pexp = "configs[?config=='system global'] | [].hostname[]"

        res = TT.jmespath_search(pexp, cnf, has_vdoms_=True, vdom="root")

        self.assertTrue(res)
        self.assertEqual(res, [val])


def _list_cnames_from_file(filepath):
    return sorted(set(line.replace("config ", '').rstrip()
                      for line in open(filepath)
                      if TT.re.match(r"^config firewall service.*", line)))


class TestCases_30_parse(C.unittest.TestCase):

    maxDiff = None
    cpaths = C.list_res_files("show_configs", "*.txt")

    def test_10_parse_show_config(self):
        for cpath in self.cpaths:
            cnf = TT.parse_show_config(cpath)
            self.assertTrue(cnf)
            TT.validate(cnf)

    def test_20_hostname_from_configs(self):
        cnf = TT.parse_show_config(self.cpaths[0])
        hostname = TT.hostname_from_configs(cnf)
        self.assertEqual(hostname, "fortigate-01")

        cnf = TT.parse_show_config(self.cpaths[1])
        hostname = TT.hostname_from_configs(cnf, has_vdoms_=True)
        self.assertEqual(hostname, "fortigate-02")

    def test_30_list_vdom_names(self):
        cnf = TT.parse_show_config(self.cpaths[0])
        self.assertEqual(TT.list_vdom_names(cnf), ["root"])

        cnf = TT.parse_show_config(self.cpaths[1])
        self.assertEqual(TT.list_vdom_names(cnf), ["ro", "root"])

    def test_50_list_cnames_for_regexp(self):
        reg = TT.re.compile(r"firewall service.*")
        for cpath in self.cpaths:
            ref = _list_cnames_from_file(cpath)
            cnf = TT.parse_show_config(cpath)
            opts = dict(has_vdoms_=TT.has_vdom(cnf))
            res = TT.list_cnames_for_regexp(cnf, reg, **opts)

            self.assertEqual(res, ref)


def houtdir(outdir, cnf):
    """Compute the output dir for each hosts.
    """
    opts = dict(has_vdoms_=TT.has_vdom(cnf))
    hostname = TT.hostname_from_configs(cnf, **opts)
    return os.path.join(outdir, hostname)


class TestCases_50(C.TestCaseWithWorkdir):

    maxDiff = None
    cpaths = C.list_res_files("show_configs", "*.txt")

    def test_10_parse_show_config_and_dump(self):
        for idx, cpath in enumerate(self.cpaths):
            outdir = TT.os.path.join(self.workdir, "{!s}".format(idx))
            cnf = TT.parse_show_config_and_dump(cpath, outdir)

            self.assertTrue(cnf)

            hdir = houtdir(self.workdir, cnf)
            for fname in (TT.METADATA_FILENAME, TT.ALL_FILENAME):
                self.assertTrue(os.path.join(hdir, fname))

    def test_20_parse_show_configs_and_dump(self):
        for _path, cnf in TT.parse_show_configs_and_dump_itr(self.cpaths,
                                                             self.workdir):
            self.assertTrue(cnf)

            hdir = houtdir(self.workdir, cnf)
            for fname in (TT.METADATA_FILENAME, TT.ALL_FILENAME):
                self.assertTrue(os.path.join(hdir, fname))

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2020 Satoru SATOH <satoru.satoh@gmail.com>.
# SPDX-License-Identifier: MIT
#
r"""Parse fortios' `show *configuration' outputs and generates various outputs.

.. versionadded:: 0.1.0

   - initial checkin
"""
from __future__ import absolute_import

import functools
import itertools
import os.path

import pandas

from . import netutils, parser, utils


ADDR_TYPES = (IP_SET, IP_NETWORK) = ("ipset", "network")

DF_ZERO = pandas.DataFrame()

ADDRS_COL_NAMES = ("addrs", "srcaddrs", "dstaddrs")

COMPRESSION_MAPS = dict(gz="gzip", bz2="bz2", zip='zip', xz='xz')


def df_by_query(path_exp, data, normalize_fn=None,
                has_vdoms_=False, vdom=None):
    """
    Make :class:`pandas.DataFrame` object from the result of JMESPath query to
    given data `data`.

    :param path_exp: JMESPath expression for search results
    :param data: A list of mapping objects
    :param has_vdoms_: True if givne `cnf` contains vdoms
    :param vdom: Specify vdom to make table

    :return: A :class:`pandas.DataFrame` object
    """
    return pandas.DataFrame(
        parser.jmespath_search(
            path_exp, data, normalize_fn=normalize_fn,
            has_vdoms_=has_vdoms_, vdom=vdom)
    )


def normalize_fa(fa_dict):
    """
    It seems that there are the following types of firewall address 'edit'
    configurations.

    - No IP address:

      {"edit": <edit_id :: int>, "uuid": ...}

    - fqdn:

      {"edit": <edit_id :: int>, "uuid": ...,
       "type": "wildcard-fqdn",
       "wildcard-fqdn": "*.google.com"}   # or "fqdn": "www.example.com"

    - iprange:

      {"edit": <edit_id :: int>, "uuid": ...,
       "type": "iprange",
       "start-ip": <ip_address>,
       "end-ip": <ip_address>}

    - unicast (host) or network address

      {"edit": <edit_id :: int>, "uuid": ...,
       "associated-interface": <interface_name>,  # 'system interface' list
       "subnet": [<ip_address>, <netmask>]}
    """
    if "start-ip" in fa_dict:  # iprange
        ipset = netutils.iprange_to_ipsets(fa_dict["start-ip"],
                                           fa_dict["end-ip"])
        fa_dict["addr_type"] = IP_SET
        fa_dict["addrs"] = ipset  # or? ' '.join(ipset)

    elif "subnet" in fa_dict:  # ip address or ip network
        ipa = netutils.subnet_to_ip(*fa_dict["subnet"])  # :: str
        if netutils.is_network_address(ipa):
            fa_dict["addr_type"] = IP_NETWORK
        else:
            fa_dict["type"] = IP_SET

        fa_dict["addrs"] = [ipa]

    else:
        fa_dict["addrs"] = []

    return fa_dict


def make_firewall_address_table_1(cnf, has_vdoms_=False, vdom=None):
    """
    :param cnf: A mapping object contains firewall configurations
    :param has_vdoms: True if givne `cnf` contains vdoms
    :param vdom: Specify vdom to make table

    :return: A :class:`pandas.DataFrame` object
    """
    rdf = df_by_query("configs[?config==`firewall address`].edits[]",
                      cnf, normalize_fn=normalize_fa,
                      has_vdoms_=has_vdoms_, vdom=vdom)
    return rdf


def make_firewall_addrgrp_table(cnf, has_vdoms_=False, vdom=None):
    """
    :param cnf: A mapping object contains firewall configurations
    :param has_vdoms: True if givne `cnf` contains vdoms
    :param vdom: Specify vdom to make table

    :return: A :class:`pandas.DataFrame` object
    """
    keys = ["edit", "uuid", "member"]
    rdf = df_by_query("configs[?config==`firewall addrgrp`].edits[]",
                      cnf, normalize_fn=None,
                      has_vdoms_=has_vdoms_, vdom=vdom
                      )
    if rdf.empty:
        return rdf

    return rdf.explode("member")[keys].reset_index(drop=True)


def make_firewall_address_table(cnf, has_vdoms_=False, vdom=None):
    """
    :param cnf: A mapping object contains firewall configurations
    :param has_vdoms: True if givne `cnf` contains vdoms
    :param vdom: Specify vdom to make table

    :return: A :class:`pandas.DataFrame` object
    """
    opts = dict(has_vdoms_=has_vdoms_, vdom=vdom)

    df_fa = make_firewall_address_table_1(cnf, **opts)
    df_ag = make_firewall_addrgrp_table(cnf, **opts)

    if df_ag.empty:
        return df_fa

    # Add columns (addr*, ...) from df_fa
    df_diff = pandas.merge(
        df_ag, df_fa, left_on="member", right_on="edit",
        suffixes=('', '_r'), how="left"
    ).drop(columns="edit_r").drop(columns="uuid_r").drop(columns="member")

    return pandas.concat([df_fa, df_diff], sort=False)


def get_firewall_address_by_edit(edit, df_fa=DF_ZERO):
    """
    :param edit: A str gives an edit ID
    :param df_fa: A :class:`pandas.DataFrame` object holds firewall addresses

    :return: A list of str gives IP addresses with prefix
    """
    return sorted(itertools.chain.from_iterable(
        e["addrs"] for e
        in df_fa[df_fa["edit"] == edit].fillna('').to_dict(orient="record")
        if e.get("addrs")
    ))


def get_firewall_address_by_edits(edits, df_fa=DF_ZERO):
    """
    :param edits: A edit or a list of edits
    :param df_fa: A :class:`pandas.DataFrame` object holds firewall addresses

    :return: A list of str gives IP addresses with prefix
    """
    if utils.is_str(edits):
        return get_firewall_address_by_edit(edits, df_fa)

    return sorted(itertools.chain.from_iterable(
        get_firewall_address_by_edit(e, df_fa=df_fa) for e in edits
    ))


def resolv_src_and_dst_in_fp(fp_dict, df_fa=DF_ZERO):
    """
    :param fp_dict: A mapping object represents a firewall policy
    :param df_fa: A :class:`pandas.DataFrame` object holds firewall addresses
    """
    fnc = functools.partial(get_firewall_address_by_edits, df_fa=df_fa)

    srcaddr = fp_dict.get("srcaddr", False)
    if srcaddr:
        fp_dict["srcaddrs"] = fnc(srcaddr)

    dstaddr = fp_dict.get("dstaddr", False)
    if dstaddr:
        fp_dict["dstaddrs"] = fnc(dstaddr)

    return fp_dict


def make_firewall_policy_table_1(cnf, df_fa, has_vdoms_=False, vdom=None):
    """
    :param cnf: A mapping object contains firewall configurations
    :param df_fa: A :class:`pandas.DataFrame` object holds firewall addresses
    :param has_vdoms: True if givne `cnf` contains vdoms
    :param vdom: Specify vdom to make table

    :return: A :class:`pandas.DataFrame` object
    """
    fa_keys = ["edit", "addr", "addrs", "comment"]  # TBD
    df_fa = df_fa.drop(columns=[k for k in df_fa.columns if k not in fa_keys])

    fnc = functools.partial(resolv_src_and_dst_in_fp, df_fa=df_fa)

    df_fp = df_by_query("configs[?config==`firewall policy`].edits[]",
                        cnf, normalize_fn=fnc,
                        has_vdoms_=has_vdoms_, vdom=vdom)

    return df_fp


def make_firewall_policy_table(cnf, vdom=None):
    """
    :param cnf: A mapping object contains firewall configurations
    :param df_fa: A :class:`pandas.DataFrame` object holds firewall addresses
    :param has_vdoms: True if givne `cnf` contains vdoms
    :param vdom: Specify vdom to make table

    :return: A :class:`pandas.DataFrame` object
    """
    opts = dict(has_vdoms_=parser.has_vdom(cnf), vdom=vdom)

    df_fa = make_firewall_address_table(cnf, **opts)
    rdf = make_firewall_policy_table_1(cnf, df_fa, **opts)

    return rdf


def guess_filetype(filepath, compression=None):
    """
    :param filepath: File path
    :param filetype: File type of `filepath`
    :param compression: Compression type
    """
    comp_exts = COMPRESSION_MAPS.keys()
    maybe_ext = os.path.splitext(filepath)[-1]

    if compression or maybe_ext.replace('.', '') in comp_exts:
        maybe_ext = os.path.splitext(os.path.splitext(filepath)[0])[-1]

    return maybe_ext.replace('.', '')


def pandas_save(rdf, outpath, filetype=None, compression=None):
    """
    :param rdf: A :class:`pandas.DataFrame` object
    :param outpath: Output file path
    :param filetype: File type to save as
    :param compression: Compression method
    """
    if filetype:
        save_fn_name = "to_" + filetype
    else:
        save_fn_name = "to_" + guess_filetype(outpath, compression=compression)
    try:
        save_fn = getattr(rdf, save_fn_name)
    except AttributeError:
        raise ValueError("Could not find appropriate save functions: "
                         "outpath={}, filetype={!s}".format(outpath, filetype))

    utils.ensure_dir_exists(outpath)
    save_fn(outpath, compression=compression)


def pandas_load(inpath, filetype=None, compression=None):
    """
    :param inpath: Output file path
    :param filetype: File type to save as
    :param compression: Compression method
    """
    load_fn_name = "read_" + (filetype if filetype else guess_filetype(inpath))
    try:
        load_fn = getattr(pandas, load_fn_name)
    except AttributeError:
        raise ValueError("Could not find appropriate load functions: "
                         "inpath={}, filetype={!s}".format(inpath, filetype))

    return load_fn(inpath, compression=compression)


def make_and_save_firewall_policy_table(cnf, outpath, vdom=None,
                                        filetype=None, compression=None):
    """
    :param cnf: A mapping object contains firewall configurations
    :param outpath: Output file path
    :param vdom: Specify vdom to make table
    :param filetype: File type to save as
    :param compression: Compression method

    :return: A :class:`pandas.DataFrame` object
    """
    rdf = make_firewall_policy_table(cnf, vdom=vdom)
    pandas_save(rdf, outpath, filetype=filetype, compression=compression)

    return rdf


def search_by_addr_1(ip_s, tbl_df, addrs_cols=ADDRS_COL_NAMES):
    """
    :param ip_s: A str represents an IP address
    :param tbl_df:
        A :class:`pandas.DataFrame` object contains ip addresses in the columns
        have one or some of `addrs_cols`.
    :param addrs_cols:
        A list of names of columns may have a set of ip addresses
    """
    if not utils.is_str(ip_s):
        raise ValueError("Expected a str but: {!r}".format(ip_s))

    ip_s = netutils.normalize_ip(ip_s)

    # TODO: I don't know how to accomplish this with pandas.DataFrame.
    rdf = tbl_df.fillna('').to_dict(orient="record")
    res = (x for x in rdf
           if any(netutils.is_ip_in_addrs(ip_s, x.get(k, []))
                  for k in addrs_cols))

    return list(res)

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2020 Satoru SATOH <satoru.satoh@gmail.com>.
# SPDX-License-Identifier: MIT
#
r"""Collect network info from fortios' configuration.
"""
from __future__ import absolute_import

import collections.abc
import ipaddress
import itertools
import logging
import os.path

import anyconfig

from . import netutils, parser, utils


NET_MAX_PREFIX = 24
NET_FILENAME = "networks.yml"
NET_ALL_FILENAME = "composed_networks.json"

NET_DATA_FMT_VER = "1.0"

# Node is an object may have some attributes.
# - id: a str or int identifies that node
# - name: a str or something describes what that node is
# - type: a str, see the definition of NODE_TYPES below
# - addrs: A list of strings each gives any IP network with prefix
# - **other optional keyword arguments
NODE_TYPES = (
    NODE_ANY, NODE_NET, NODE_HOST, NODE_ROUTER, NODE_SWITCH, NODE_FIREWALL
) = (
    "any", "network", "host", "router", "switch", "firewall"
)

LOGGER = logging.getLogger(__name__)


def add_node_info(node):
    """
    :param node: A mapping object denotes the network / host node
    :return: The mapping object with additional info
    """
    node["type_id"] = NODE_TYPES.index(node["type"])
    node["class"] = "node {0[type]}".format(node)
    node["label"] = "{0[name]} ({0[type]})".format(node)

    return node


def make_net_node(net, extra_info=True):
    """
    :param net: A ipaddress.IPv*Network object
    :return: A mapping object represents the network node
    """
    net_s = str(net)
    node = dict(id=net_s, name=net_s, type=NODE_NET, addrs=[net_s])

    if extra_info:
        node = add_node_info(node)

    return node


def make_edge(nodes, distance):
    """
    :param nodes: A tuple of address strings
    :param distance: 'Distance' between edges

    :return: A mapping object will be used in D3.js
    """
    name = "{}_{}".format(*nodes)

    return dict(type="edge", distance=distance, id=name,
                source=nodes[0], target=nodes[1])


def list_interface_addrs(cnf, **sargs):
    """
    Get a list of interface addresses from interface configuration data.

    :param cnf: A mapping object holding fortios configuration data
    :param sargs: Keyword argument will be passed to parser.jmespath_search

    :return: A list of ipaddress.IPv*Interface objects give interface addresses
    """
    query = "configs[?config=='system interface'].edits[].ip"
    qres = parser.jmespath_search(query, cnf, **sargs) or []

    return [ipaddress.ip_interface("{}/{}".format(*ip)) for ip in qres]


def list_firewall_addrs(cnf, **sargs):
    """
    Get a list of network addresses from firewall address configuration data.

    :param cnf: A mapping object holding fortios configuration data
    :param sargs: Keyword argument will be passed to parser.jmespath_search

    :yield: A str gives a network address
    """
    # .. note:: I don't know why it doesn't work as expected always.
    # query = "configs[?config=='firewall address'].edits[][?subnet].subnet"
    query = "configs[?config=='firewall address'].edits[]"
    qres = parser.jmespath_search(query, cnf, **sargs)
    if not qres:
        return []

    sns = [netutils.subnet_to_ip(*x["subnet"]) for x in qres if "subnet" in x]
    irs = list(itertools.chain.from_iterable(
        netutils.iprange_to_ipsets(x["start-ip"], x["end-ip"])
        for x in qres if x.get("type") == "iprange"
    ))
    return list(set(sns + irs))


def _node_and_edges_from_fa_networks_itr(inets, nets):
    """
    :param inets:
        A list of network addresses with prefix, e.g. 10.0.1.0/24, of which
        network is linked to the interface
    :param networks:
        A list of network addresses with prefix same as the above but came from
        the other configuration data such like firewall address list
    """
    for net in nets:
        inet = netutils.find_nearest_network(net, inets)
        if inet == net:
            continue

        distance = netutils.distance(net, inet)
        if distance == netutils.math.inf:
            distance = 32 * 2  # Avoid JSON syntax error by math.inf.

        yield make_net_node(net)
        yield make_edge((inet, net), distance)


def node_and_edges_from_config_file_itr(filepath, prefix=NET_MAX_PREFIX):
    """
    Get a node and edges (node and network links) information from fortigate's
    parsed configuration file.

    :param filepath: Path to the JSON file contains fortigate's configurations
    :param prefix: 'Largest' network prefix to find

    :return: A graph data
    :raises: ValueError
    """
    cnf = parser.load(filepath)
    if not cnf:
        raise ValueError("Something goes wrong with the input: "
                         "{}".format(filepath))

    opts = dict(has_vdoms_=parser.has_vdom(cnf))

    hostname = parser.hostname_from_configs(cnf, **opts)
    if not hostname:
        raise ValueError("Hostname configuration was not found. "
                         "Is this valid config data?: "
                         "{}".format(filepath))

    ifas = list_interface_addrs(cnf, **opts)  # [<ip_addr_str_w_prefix>]
    if not ifas:
        raise ValueError("Interfaces were not found or ip address "
                         "is not set. Check the configuration data: "
                         "{}".format(filepath))

    host = add_node_info(dict(id=hostname, name=hostname, type=NODE_FIREWALL,
                              addrs=[str(i) for i in ifas]))
    yield host  # host node

    ifns = [i.network for i in ifas]  # :: [IPv4Network]
    for ifn in ifns:
        yield make_net_node(ifn)  # (network) node
        yield make_edge((hostname, str(ifn)), 1)

    inets = [str(i) for i in ifns]  # :: [str]
    fas = list_firewall_addrs(cnf, **opts)  # :: [str]

    # networks connected from the interfaces
    cnets = [a for a in fas
             if (netutils.is_network_address(a) and
                 a not in inets and
                 a != "0.0.0.0/32")]
    for obj in _node_and_edges_from_fa_networks_itr(inets, cnets):
        yield obj

    # compute networks contains the hosts not in the previous `cnets` connected
    # from the interfaces.
    cnets = set(str(ipaddress.ip_network(a).supernet(new_prefix=prefix))
                for a in fas
                if (a not in cnets + inets and
                    not netutils.is_ip_in_addrs(a, cnets)))

    for obj in _node_and_edges_from_fa_networks_itr(inets, cnets):
        yield obj


def make_and_save_networks_from_config_file(filepath, outpath=None,
                                            prefix=NET_MAX_PREFIX):
    """
    Make a graph of netwrks of nodes and edges (node and network links)
    information from fortigate's parsed configuration file.

    :param filepath: Path to the JSON file contains fortigate's configurations
    :param prefix: 'Largest' network prefix to find

    :return: A graph data contains metadata, nodes and links data
    :raises: ValueError
    """
    graph = list(node_and_edges_from_config_file_itr(filepath, prefix=prefix))
    nodes = [x for x in graph if x["type"] != "edge"]
    edges = [x for x in graph if x["type"] == "edge"]

    if not outpath:
        outpath = os.path.join(os.path.dirname(filepath), NET_FILENAME)

    metadata = dict(type="metadata", input=filepath, prefix=prefix,
                    timestamp=utils.timestamp(), version=NET_DATA_FMT_VER)
    res = dict(metadata=metadata, nodes=nodes, links=edges)

    utils.ensure_dir_exists(outpath)
    anyconfig.dump(res, outpath)

    return res


def make_and_save_networks_from_config_files_itr(filepaths,
                                                 prefix=NET_MAX_PREFIX):
    """
    Similar to :func:`make_and_save_networks_from_config_file` but allow giving
    multiple file paths.

    :param filepaths:
        A list of paths to the JSON files contains the parsed results of
        fortigate's 'show configuration' outputs
    :param prefix: 'Largest' network prefix to find

    :yield:
        A tuple of (input file path, graph data contains metadata, nodes and
        links data)
    """
    for fpath in filepaths:
        yield (fpath,
               make_and_save_networks_from_config_file(fpath, prefix=prefix))


def load_network_graph_files_itr(filepaths):
    """
    Load netowrk graph data

    :param filepaths:
        A list of path to the JSON file contains network graph data

    :return: A graph data contains metadata, nodes and links data
    """
    for filepath in filepaths:
        graph = anyconfig.load(filepath)

        if not isinstance(graph, collections.abc.Mapping) or \
                "nodes" not in graph or "links" not in graph:
            LOGGER.warning("Not look network graph data: %s", filepath)
            continue

        yield graph["metadata"]

        for node in graph["nodes"]:
            yield node

        for link in graph["links"]:
            yield link


def _compose_nodes_itr(graph):
    """Compose nodes.
    """
    # select unique nodes
    nodes = {n["id"]: n for n in graph
             if n["type"] not in ("edge", "metadata")}.values()

    for node in nodes:
        node["class"] = "node {type}".format(**node)
        node["label"] = "{id} {type}".format(**node)

        yield node


def compose_network_graph_files(filepaths, outpath=None):
    """
    Make a graph of networks of nodes and edges (node and network links)
    information from fortigate's parsed configuration file.

    :param filepaths:
        A list of path to the JSON file contains network graph data
    :param outpath: Output file path

    :return: A graph data contains metadata, nodes and links data
    """
    (nit, lit, mit) = itertools.tee(load_network_graph_files_itr(filepaths), 3)

    nodes = list(_compose_nodes_itr(nit))
    links = list({link["id"]: link for link in lit
                 if link["type"] == "edge"}.values())

    metadata = dict(inputs=[m["input"] for m in mit
                            if m["type"] == "metadata"],
                    timestamp=utils.timestamp(), version=NET_DATA_FMT_VER)

    res = dict(metadata=metadata, nodes=nodes, links=links)

    if not outpath:
        outpath = os.path.join(os.path.dirname(filepaths[0]), NET_ALL_FILENAME)

    utils.ensure_dir_exists(outpath)
    anyconfig.dump(res, outpath)

    return res

# vim:sw=4:ts=4:et:



"""setup.py to build packages.
"""
from __future__ import absolute_import

import glob
import os
import re
import setuptools
import setuptools.command.bdist_rpm


VERSION = False
for pyf in glob.glob("src/*/__init__.py"):
    matches = [m.groups() for m in (re.match(r'__version__ = "([0-9.]+)"', l)
                                    for l in open(pyf)) if m]
    if matches:
        VERSION = matches[0][0]
        break

assert VERSION

# For daily snapshot versioning mode:
RELEASE = "1%{?dist}"
if os.environ.get("_SNAPSHOT_BUILD", None) is not None:
    import datetime
    RELEASE = datetime.datetime.now().strftime("%Y%m%d%H%M%S")


def _replace(line):
    """Replace some strings in the RPM SPEC template"""
    if "@VERSION@" in line:
        return line.replace("@VERSION@", VERSION)

    if "@RELEASE@" in line:
        return line.replace("@RELEASE@", RELEASE)

    if "Source0:" in line:  # Dirty hack
        return "Source0: %{pkgname}-%{version}.tar.gz"

    return line


# pylint: disable=invalid-name
class bdist_rpm(setuptools.command.bdist_rpm.bdist_rpm):
    """Override the default content of the RPM SPEC.
    """
    spec_tmpl = os.path.join(os.path.abspath(os.curdir),
                             "pkg/package.spec.in")

    def _make_spec_file(self):
        return [_replace(l.rstrip()) for l in open(self.spec_tmpl)]


setuptools.setup(version=VERSION,
                 package_dir={'': 'src'},
                 cmdclass=dict(bdist_rpm=bdist_rpm))

# vim:sw=4:ts=4:et:
